/* The Merry Turnip - TODO: Description
 * Copyright © DAVY Guillaume 2009 <davyg2@gmail.com>
 * 
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

// TODO
/*!	\file coeur.cpp
	\brief Implement coeur class
	
	Contain all member of class coeur.
	
	\todo Remplacer les vector path par des listes indexé de la STL
	\todo Creer repertoir log
	\todo Use the event system to launch personal event(like changing chap, animation, script,...)
	\todo Add in the xml a line to say the path of the texture folder and ad it to irrlicht filesytem
	\bug Not respond when launched fullscreen
*/

#include <game.h>
#include <startmenu.h>

/*!
	\class coeur
	This class contain all the reference to the irrlicht and irrklang engine
	It contains most of the stuff needed by other class. It's the only class
	who have to be create, all th others are handle by this one, and the xml config file.
	This is class is configured by two xml files : the config.xml for all basic stuff
	and the main.xml which is the main config filoe of the game\n
	If you want to know more about the xml configuration see config.xml and main.xml
	
	\file config.xml
	\ingroup xml
	\section Description
	This file is used by two class : the startMenu, in order to let the user changed the
	configuration and the coeur. It contains the irrlicht configuration, the sound configuration,
	the language configuration and the patch to the chosen game and the startmenu\n
	\n
	\section Xml
	I assume that you know how general xml are formated(tags system like html)\n
	\n
	The xml should begin whith a line like that :
	\code <?xml version="1.0"?>	\endcode \n
	There is 6 types of element possible :
		- \e config : this is the main element it takes no attribute
		
		\code <config> \endcode
		
		- \e device : Config the device
			- \e driver : Number of the graphic driver to use
				- \e 0 : No Graphic driver
				- \e 1 : Software driver(Built in)
				- \e 2 : Burnings video driver
				- \e 3 : Directx 8.0 driver(Windows only)
				- \e 4 : Directx 9.0 driver(Windows only)
				- \e 5 : OPENGL
			- \e bits : Size of color bits(16 would be fine)
			- \e fs : Does the game is fullscreen
			- \e sb : Does the stencil shadow are activated
			- \e vs : Does the vertical sync are activated
		
		\code <device driver="5" bits="16" fs="0" sb="1" vs="0"/> \endcode
		
		- \e dimension : Dimension of the screen
			- \e width : Width of the screen
			- \e height : Height of the screen
		
		\code <dimension width="1024" height="768" /> \endcode
		
		- \e %main : Reference to the game file
			- \e file : Path of the xml game file
		
		\code <main file="data/demo/main.xml" /> \endcode
		
		- \e %local : Language selection
			- \e lang : Number of the language
				- \e 0 : French
		
		\code <local lang="0" /> \endcode
		
		- \e %startmenu : Reference to the startmenu file
			- \e file : Path of the xml startmenu file
		
		\code <startmenu file="data/startmenu/startmenu.xml" /> \endcode
	\n
	
	\section Example
	
	To see a complete example just look at the source code of this file\n
	The example contain default configuration, should easy to understand
*/

//! Construct the coeur class, call base::base() and create all the class and initialize some value
coeur::coeur() : base("coeur")
{
	py = new pyInt;
	in = new interface;
	co = new console;
	mycredit = new credit;

	currentChap = new chap;
	currentLieu = new lieu;
	currentLocal = new local;

	mylogger = new logger;

	pathChap = new std::vector<stringc*>;
	pathLieu = new std::vector<stringc*>;
	pathLocal = new std::vector<stringc*>;

	dr = EDT_NULL;
	dim = dimension2d<s32>(640, 480);
	bits = 8;
	fullscreen = false;
	stencilbuffer = false;
	vsync = false;

	lang = 0;
}

//! Destruct the coeur class, delete all what is needed
coeur::~coeur()
{
	delete pathLocal;
	delete pathLieu;
	delete pathChap;

	delete currentLocal;
	delete currentLieu;
	delete currentChap;

	delete mylogger;

	delete mycredit;
	delete co;
	delete in;
	delete py;
}

/** Call base::init() and read xmlChap 
	\param startmenu pointer to the startmenu to update the progress bar
	\return 0 if all goes well*/
int coeur::init(startmenu* cb)
{
	//make the progress bar go on
	cb->progress();
	
	bool classic;

	stringc pathMain = "";
	
	stringc pathConfig = "config.xml";

	IXMLReader* xmlConfig = cb->getFileSystem()->createXMLReader(pathConfig.c_str());

	cb->progress();
	while(xmlConfig && xmlConfig->read())
	{
		if (xmlConfig->getNodeType()==EXN_ELEMENT)
		{
			if (core::stringw("device") == xmlConfig->getNodeName())
			{
				dr = E_DRIVER_TYPE(xmlConfig->getAttributeValueAsInt(L"driver"));
				bits = xmlConfig->getAttributeValueAsInt(L"bits");
				fullscreen = xmlConfig->getAttributeValueAsInt(L"fs");
				stencilbuffer = xmlConfig->getAttributeValueAsInt(L"sb");
				vsync = xmlConfig->getAttributeValueAsInt(L"vs");
			}
			else if (core::stringw("dimension") == xmlConfig->getNodeName())
			{
				dim = dimension2d<s32>(xmlConfig->getAttributeValueAsInt(L"width"),xmlConfig->getAttributeValueAsInt(L"height"));
			}
			else if (core::stringw("local") == xmlConfig->getNodeName())
			{
				lang = xmlConfig->getAttributeValueAsInt(L"lang");
			}
			else if (core::stringw("main") == xmlConfig->getNodeName())
			{
				pathMain = xmlConfig->getAttributeValue(L"file");
			}
			else if (core::stringw("interface") == xmlConfig->getNodeName())
			{
				classic = xmlConfig->getAttributeValueAsInt(L"classic")==1;
			}
		}
	}

	xmlConfig->drop();

	device = createDevice(dr,dim,bits,fullscreen,stencilbuffer,vsync);
	
	#ifdef AUDIO
	engine = createIrrKlangDevice();
	#endif
	
	driver = device->getVideoDriver();
	smgr = device->getSceneManager();
	env = device->getGUIEnvironment();
	
	mylogger->init(this);
	base::init(0,this);

	log(stringc("Main xml : ") + pathMain + " loading ...");
	
	pathGame = device->getFileSystem()->getAbsolutePath(device->getFileSystem()->getFileDir(pathMain)).c_str();
	device->getFileSystem()->addFolderFileArchive(pathGame.c_str());
	
	IXMLReader* xmlMain = device->getFileSystem()->createXMLReader(pathMain.c_str());

	stringc pathInterface = "";
	stringc pathCredit = "";
	
	cb->progress();
	while(xmlMain && xmlMain->read())
	{
		if (xmlMain->getNodeType()==EXN_ELEMENT)
		{
			if (core::stringw("main") == xmlMain->getNodeName())
			{
				pathChap->reserve(xmlMain->getAttributeValueAsInt(L"nbrchap"));
				pathLieu->reserve(xmlMain->getAttributeValueAsInt(L"nbrlieu"));
				pathLocal->reserve(xmlMain->getAttributeValueAsInt(L"nbrlocal"));
			}
			else if (core::stringw("interface") == xmlMain->getNodeName())
			{
				pathInterface = xmlMain->getAttributeValue(L"file");
				log( stringc("pathInterface : ") + pathInterface.c_str());
			}
			else if (core::stringw("chap") == xmlMain->getNodeName())
			{
				pathChap->push_back(new stringc(xmlMain->getAttributeValue(L"file")));
				log( stringc("pathChap : ") + pathInterface.c_str());
			}
			else if (core::stringw("lieu") == xmlMain->getNodeName())
			{
				pathLieu->push_back(new stringc(xmlMain->getAttributeValue(L"file")));
				log( stringc("pathLieu : ") + (*pathLieu)[pathLieu->size()-1]->c_str());
			}
			else if (core::stringw("local") == xmlMain->getNodeName())
			{
				pathLocal->push_back(new stringc(xmlMain->getAttributeValue(L"file")));
				log( stringc("pathLocal : ") + (*pathLocal)[pathLocal->size()-1]->c_str());
			}
			else if (core::stringw("credit") == xmlMain->getNodeName())
			{
				pathCredit = xmlMain->getAttributeValue(L"file");
				log( stringc("pathCredit : ") + pathCredit.c_str() );
			}
		}
	}

	xmlMain->drop();

	cb->progress();
	metaSelector = smgr->createMetaTriangleSelector();
	
	cam = smgr->addCameraSceneNodeFPS();
	cam->setPosition(vector3df(0,150,-200));
	device->getCursorControl()->setVisible(false);
	log("Camera ajoute");
	
	q = false;

	cb->progress();
	IXMLReader* xmlInterface = device->getFileSystem()->createXMLReader(pathInterface.c_str());
	if (xmlInterface)
	{
		in->init(this,env,xmlInterface,classic);
		xmlInterface->drop();
		log("Interface initialized");
	}
	else
	{
		log(stringc("Problem loading interface xml :") + pathInterface + ", exiting...");
		return 1;
	}
	cb->progress();
	IXMLReader* xmlCredit = device->getFileSystem()->createXMLReader(pathCredit.c_str());
	mycredit->init(this,env,xmlCredit);
	xmlCredit->drop();
	log("Credit initialized");

	cb->progress();
	loadChap(0);
	log("Chapitre load");
	
	cb->progress();
	loadLocal(lang);
	log("Local load");

	setCaption(currentLocal->getCaption() + " - " + VERSION);
	
	py->init(this);
	log("Python initialized");
	
	cb->progress();
	co->init(this,env);
	log("Console initialized");

	cb->progress();
	device->setEventReceiver(this);
	log("Event receiver on");

	ISceneNodeAnimator* anim = smgr->createCollisionResponseAnimator(metaSelector, cam, vector3df(10,40,8), vector3df(0,-980,0), vector3df(0,40,0));
	cam->addAnimator(anim);
	anim->drop();
	
	selectedAction=0;
	
	smgr->setShadowColor(video::SColor(150,0,0,0));
	
	activeScene = true;
	activeIn = true;
	
	return 0;
}

ISceneNode* coeur::getSelectedNode()
{
	line3df line = smgr->getSceneCollisionManager()->getRayFromScreenCoordinates(curPos);
		
	ISceneNode* best = 0;
	f32 dist = 1000000;
		
	pickNode(smgr->getRootSceneNode(), line, dist, best);
	
	return best;	
}

void coeur::pickNode(ISceneNode* root,line3df ray,f32& outbestdistance,ISceneNode*& outbestnode)
{
	const core::list<ISceneNode*>& children = root->getChildren();
	core::list<ISceneNode*>::ConstIterator it = children.begin();
	for (; it != children.end(); ++it)
	{
		ISceneNode* current = *it;

		if (current->isVisible() && !current->isDebugObject() && current->getID() != -1)
		{
			vector3df p;
			triangle3df t;
			if (smgr->getSceneCollisionManager()->getCollisionPoint(ray,current->getTriangleSelector(),p,t))
			{
				f32 distance = p.getDistanceFrom(ray.start);
				if (distance < outbestdistance)
				{
					outbestnode = current;
					outbestdistance = distance;
				}
			}
		}
		pickNode(current, ray, outbestdistance, outbestnode);
	}
}

/** Desc
	\return */
int coeur::update()
{
	int r = device->run();
	if (r && device->isWindowActive())
	{
		curPos=device->getCursorControl()->getPosition();
		
		selectedAction = getActionMesh(getSelectedNode());
		
		if (activeIn)
			in->update();
		co->update();
		mycredit->update();
		driver->beginScene(true, true, SColor(0,0,0,0));
		
		if (activeScene)
			smgr->drawAll();
		env->drawAll();
		
		driver->endScene();
		if (q)
		{
			log("Quitting");
			return 0;
		}
	}
	return r;
}

int coeur::close()
{
	co->close();
	currentLocal->close();
	currentLieu->close();
	currentChap->close();
	in->close();
	mycredit->close();
	log("All subclass closed");
	
	while (! pathChap->empty())
	{
		log(stringc("Deleting last pathChap"));
		delete pathChap->back();
		pathChap->pop_back();
	}
	while (! pathLieu->empty())
	{
		log(stringc("Deleting last pathLieu"));
		delete pathLieu->back();
		pathLieu->pop_back();
	}
	while (! pathLocal->empty())
	{
		log(stringc("Deleting last pathChap"));
		delete pathLocal->back();
		pathLocal->pop_back();
	}
	log("All path have been deleted");
	
	py->close();
	
	log("coeur is going to close itself and the logger");
	log("Irrlicht device and irrklang engine will be drop too");
	base::close();
	mylogger->close();
	
	device->drop();
	#ifdef AUDIO
	engine->drop();
	#endif
	return 1;
}

int coeur::setCaption(stringw caption)
{
	device->setWindowCaption(caption.c_str());
	log(stringc("Change window caption to ") + caption.c_str());
	return 1;
}

IAnimatedMeshSceneNode* coeur::addMesh(stringc pathm)
{
	IAnimatedMesh* mesh = smgr->getMesh(pathm.c_str());
	if (!mesh)
		return 0;

	IAnimatedMeshSceneNode* node = smgr->addAnimatedMeshSceneNode(mesh);
	node->setFrameLoop(0, 0);
	node->setLoopMode(false);
	
	ITriangleSelector* selector = smgr->createOctTreeTriangleSelector(mesh, node);
	node->setTriangleSelector(selector);
	metaSelector->addTriangleSelector(selector);
	selector->drop();

	node->setMaterialFlag(EMF_NORMALIZE_NORMALS, true);
	node->addShadowVolumeSceneNode();

	if (!node)
		return 0;
	log(stringc("New mesh added : \"") + pathm + "\" number of frame :" + stringc(node->getEndFrame()));
	return node;
}

ILightSceneNode* coeur::addLight(int radius)
{
	ILightSceneNode* light = smgr->addLightSceneNode(0,vector3df(0,0,0),SColorf(0,0,0,0),radius);
	//smgr->addCubeSceneNode(5,light);
	return light;
}

ITexture* coeur::getTexture(stringc path)
{
	log(stringc("Get texture ") + path);
	return driver->getTexture(path.c_str());
}

dimension2d<s32> coeur::getDim()
{
	return dim;
}

actionMesh *coeur::getSelectedAction()
{
	return selectedAction;
}

position2d<s32> coeur::getCurPos()
{
	return curPos;
}

u32 coeur::getTime()
{
	return device->getTimer()->getRealTime();
}


int coeur::play(stringc path)
{
	log(stringc("Play sound ") + path);
#ifdef AUDIO
	engine->play2D((pathGame + "/" + path).c_str(), false);
#endif
	return 1;
}

bool coeur::OnEvent(const SEvent& event)
{
	if (co->OnEvent(event))
		return false;
	if (mycredit->OnEvent(event))
		return false;
	if (event.EventType==EET_KEY_INPUT_EVENT)
	{
		if (event.KeyInput.PressedDown)
		{
			switch(event.KeyInput.Key)
			{
			case KEY_KEY_Q:
				q = true;
				return true;
			case KEY_KEY_G:
				startCredit();
				return true;
			case KEY_KEY_M:
				cam->setInputReceiverEnabled(false);
				return true;
			default:
				break;
			}
		}
		else if (event.KeyInput.Key == KEY_KEY_M)
		{
			cam->setInputReceiverEnabled(true);
			return true;
		}
	}
	if (in->OnEvent(event))
		return true;
	return false;
}

int coeur::loadChap(int i)
{
	if (currentChap->isInit())
		currentChap->close();
	IXMLReader* xmlChap = device->getFileSystem()->createXMLReader(pathChap->at(i)->c_str());
	int numLieu = currentChap->init(i,this,xmlChap);
	xmlChap->drop();

	loadLieu(numLieu);
	
	log(stringc("Chap ") + stringc(i) + " load");

	return 1;
}

int coeur::loadLieu(int i)
{
	if (currentLieu->isInit())
		currentLieu->close();
	IXMLReader* xmlLieu = device->getFileSystem()->createXMLReader(pathLieu->at(i)->c_str());
	currentLieu->init(i,this,xmlLieu);
	xmlLieu->drop();

	log(stringc("Lieu ") + stringc(i) + " load");

	return 1;
}

int coeur::loadLocal(int i)
{
	if (currentLocal->isInit())
		currentLocal->close();
	IXMLReader* xmlLocal = device->getFileSystem()->createXMLReader(pathLocal->at(i)->c_str());
	currentLocal->init(i,this,xmlLocal);
	xmlLocal->drop();

	log(stringc("Local ") + stringc(i) + " load");

	return 1;
}

int coeur::startCredit()
{
	cam->setInputReceiverEnabled(false);
	activeScene = false;
	activeIn = false;
	mycredit->start();
	return 1;
}

item coeur::getItem(int id)
{
	return currentChap->getItemFromId(id);
}

//! /todo Maybe returning a pointer is not a good idea
actionMesh* coeur::getActionMesh(int id)
{
	return currentLieu->getActionMesh(id);
}

actionMesh* coeur::getActionMesh(ISceneNode * node)
{
	return currentLieu->getActionMesh(node);
}

localItem* coeur::getObservationMesh(int id)
{
	return  currentLocal->getObservationMesh(id);
}

localItem* coeur::getObservationItem(int id)
{
	return  currentLocal->getObservationItem(id);
}

localItem* coeur::getCombinaison()
{
	return  currentLocal->getCombinaison();
}

logger* coeur::getLogger()
{
	return mylogger;
}

stringc coeur::getVersion()
{
	return stringc(VERSION);
}

int coeur::anim(int id,int begin,int end)
{
	IAnimatedMeshSceneNode* node = getActionMesh(id)->getNode();
	
	if (!node)
		return -1;
	
	node->setFrameLoop(begin,end);
	
	metaSelector->removeTriangleSelector(node->getTriangleSelector());
	
	ITriangleSelector* selector = smgr->createOctTreeTriangleSelector(node->getMesh()->getMesh(40), node);
	node->setTriangleSelector(selector);
	metaSelector->addTriangleSelector(selector);
	selector->drop();
	
	
	return 1;
}

int coeur::setSkyBox(stringc top,stringc bottom,stringc left,stringc right,stringc front,stringc back)
{
	smgr->addSkyBoxSceneNode (getTexture(top),getTexture(bottom),getTexture(left),getTexture(right),getTexture(front),getTexture(back));
	return 1;
}


IParticleSystemSceneNode* coeur::addParticleSystem()
{
	return smgr->addParticleSystemSceneNode(false);
}





